import functools


class DCORLogs:
    def __init__(self, api):
        self.api = api

    def __getitem__(self, key):
        return self._logs[key]

    def __len__(self):
        return len(self._logs)

    def keys(self):
        return self._logs.keys()

    @property
    @functools.lru_cache()
    def _logs(self):
        return self.api.get(query="logs")



"""
Basins are other .rtdc files on disk or online (DCOR, S3) which originate
from the same dataset but (potentially) contain more features. Basins
are useful if you would like to have a small copy of selected
features in a separate file while still being able to access
all features from the original file. E.g. you could load a small
.rtdc file from your local disk and access the larger "image"
feature data from an S3 basin. Basins are active by default, which
means that if you load a dataset that defines basins and these
basins are available, they will be integrated seamlessly into your
analysis pipeline. You can find out which features originate from
other basins via the ``features_basin`` property of an
:class:`.RTDCBase` instance.
"""
from __future__ import annotations

import abc


class Basin(abc.ABC):
    """A basin represents data from an external source

    The external data must be a valid RT-DC dataset, subclasses
    should ensure that the corresponding API is available.
    """
    id_getters = {}

    def __init__(self, location, name=None, description=None,
                 features=None, measurement_identifier=None, **kwargs):
        """

        Parameters
        ----------
        location: str
            Location of the basin, this can be a path or a URL, depending
            on the implementation of the subclass
        name: str
            Human-readable name of the basin
        description: str
            Lengthy descrition of the basin
        features: list of str
            List of features this basin provides; This list is enforced,
            even if the basin actually contains more features.
        measurement_identifier: str
            A measurement identifier against which to check the basin.
            If this is set to None (default), there is no certainty
            that the downstream dataset is from the same measurement.
        """
        #: location of the basin (e.g. path or URL)
        self.location = location
        #: user-defined name of the basin
        self.name = name
        #: lengthy description of the basin
        self.description = description
        # features this basin provides
        self._features = features
        #: measurement identifier of the referencing dataset
        self.measurement_identifier = measurement_identifier
        self._measurement_identifier_verified = False
        #: additional keyword arguments passed to the basin
        self.kwargs = kwargs
        self._ds = None

    def _assert_measurement_identifier(self):
        """Make sure the basin matches the measurement identifier

        This method caches its result, i.e. only the first call is slow.
        """
        if not self._measurement_identifier_verified:
            if self.measurement_identifier is None:
                self._measurement_identifier_verified = True
            else:
                self._measurement_identifier_verified = (
                    self.get_measurement_identifier()
                    == self.measurement_identifier)
        if not self._measurement_identifier_verified:
            raise KeyError(f"Measurement identifier of {self.ds} "
                           f"({self.get_measurement_identifier()}) does "
                           f"not match {self.measurement_identifier}!")

    @property
    @abc.abstractmethod
    def basin_format(self):
        """Basin format (:class:`.RTDCBase` subclass), e.g. "hdf5" or "s3"
        """
        # to be implemented in subclasses

    @property
    @abc.abstractmethod
    def basin_type(self):
        """Storage type to use (e.g. "file" or "remote")"""
        # to be implemented in subclasses

    @property
    def ds(self):
        """The :class:`.RTDCBase` instance represented by the basin"""
        if self._ds is None:
            if not self.is_available():
                raise ValueError(f"Basin {self} is not available!")
            self._ds = self.load_dataset(self.location, **self.kwargs)
        return self._ds

    @property
    def features(self):
        """Features made available by the basin"""
        if self._features is None:
            self._features = self.ds.features_innate
        return self._features

    def as_dict(self):
        """Return basin kwargs for :func:`RTDCWriter.store_basin`

        Note that each subclass of :class:`.RTDCBase` has its own
        implementation of :func:`.RTDCBase.basins_get_dicts` which
        returns a list of basin dictionaries that are used to
        instantiate the basins in :func:`RTDCBase.basins_enable`.
        This method here is only intended for usage with
        :func:`RTDCWriter.store_basin`.
        """
        return {
            "basin_name": self.name,
            "basin_type": self.basin_type,
            "basin_format": self.basin_format,
            "basin_locs": [self.location],
            "basin_descr": self.description,
            "basin_feats": self.features,
        }

    def get_feature_data(self, feat):
        """Return an object representing feature data of the basin"""
        self._assert_measurement_identifier()
        return self.ds[feat]

    def get_measurement_identifier(self):
        """Return the identifier of the dataset"""
        return self.ds.get_measurement_identifier()

    @abc.abstractmethod
    def is_available(self):
        """Return True if the basin is available"""

    @abc.abstractmethod
    def load_dataset(self, location, **kwargs):
        """Subclasses should return an instance of :class:`.RTDCBase`"""


def get_basin_classes():
    bc = {}
    for bcls in Basin.__subclasses__():
        if hasattr(bcls, "basin_format"):
            bc[bcls.basin_format] = bcls
    return bc



import functools
import re
import socket
from urllib.parse import urlparse


try:
    import s3fs
except ModuleNotFoundError:
    S3FS_AVAILABLE = False
else:
    S3FS_AVAILABLE = True


from .feat_basin import Basin

from .fmt_hdf5 import RTDC_HDF5


#: Regular expression for matching a DCOR resource URL
REGEXP_S3_URL = re.compile(
    r"^(https?:\/\/)"  # protocol (http or https or omitted)
    r"([a-z0-9-\.]*)(\:[0-9]*)?\/"  # host:port
    r".+\/"  # bucket
    r".+"  # key
)


class RTDC_S3(RTDC_HDF5):
    def __init__(self,
                 url: str,
                 secret_id: str = "",
                 secret_key: str = "",
                 *args, **kwargs):
        """Access RT-DC measurements in an S3-compatible object store

        This is essentially just a wrapper around :class:`.RTDC_HDF5`
        with `s3fs` passing a file object to h5py.

        Parameters
        ----------
        url: str
            Full URL to an object in an S3 instance
        secret_id: str
            S3 access identifier
        secret_key: str
            Secret S3 access key
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: str
            The URL to the object
        """
        if not S3FS_AVAILABLE:
            raise ModuleNotFoundError(
                "Package `s3fs` required for S3 format!")
        s3fskw = get_s3fs_kwargs(url=url,
                                 secret_id=secret_id,
                                 secret_key=secret_key)
        _, s3_path = parse_s3_url(url)
        self._fs = s3fs.S3FileSystem(**s3fskw)
        self._f3d = self._fs.open(s3_path, mode='rb')
        # Initialize the HDF5 dataset
        super(RTDC_S3, self).__init__(
            h5path=self._f3d,
            *args,
            **kwargs)
        # Override self.path with the actual S3 URL
        self.path = url


class S3Basin(Basin):
    basin_format = "s3"
    basin_type = "remote"

    def __init__(self, *args, **kwargs):
        self._available_verified = False
        super(S3Basin, self).__init__(*args, **kwargs)

    def load_dataset(self, location, **kwargs):
        return RTDC_S3(location, enable_basins=False, **kwargs)

    def is_available(self):
        """Check for s3fs and object availability

        Caching policy: Once this method returns True, it will always
        return True.
        """
        if not self._available_verified:
            self._available_verified = (
                    S3FS_AVAILABLE and is_s3_object_available(self.location))
        return self._available_verified


@functools.lru_cache()
def get_s3fs_kwargs(url: str,
                    secret_id: str = "",
                    secret_key: str = "",
                    ):
    """Return keyword arguments for s3fs

    Parameters
    ----------
    url: str
        full URL to the object
    secret_id: str
        S3 access identifier
    secret_key: str
        Secret S3 access key
    """
    s3_endpoint, _ = parse_s3_url(url)
    s3fskw = {
        "endpoint_url": s3_endpoint,
        # A large block size makes loading metadata really slow.
        "default_block_size": 2**18,
    }
    if secret_id and secret_key:
        # We have an id-key pair.
        s3fskw["key"] = secret_id
        s3fskw["secret"] = secret_key
        s3fskw["anon"] = False  # this is the default
    else:
        # Anonymous access has to be enabled explicitly.
        # Normally, s3fs would check for credentials in
        # environment variables and does not fall back to
        # anonymous use.
        s3fskw["anon"] = True
    return s3fskw


@functools.lru_cache()
def is_s3_object_available(url: str,
                           secret_id: str = "",
                           secret_key: str = "",
                           ):
    """Check whether an S3 object is available

    Parameters
    ----------
    url: str
        full URL to the object
    secret_id: str
        S3 access identifier
    secret_key: str
        Secret S3 access key
    """
    avail = False
    if is_s3_url(url):
        urlp = urlparse(url)
        # default to https if no scheme or port is specified
        port = urlp.port or (80 if urlp.scheme == "http" else 443)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            # Try to connect to the host
            try:
                s.connect((urlp.netloc, port))
            except (socket.gaierror, OSError):
                pass
            else:
                # Try to access the object
                s3fskw = get_s3fs_kwargs(url=url,
                                         secret_id=secret_id,
                                         secret_key=secret_key)
                _, s3_path = parse_s3_url(url)
                fs = s3fs.S3FileSystem(**s3fskw)
                try:
                    avail = fs.exists(s3_path)
                except OSError:
                    pass
    return avail


@functools.lru_cache()
def is_s3_url(string):
    """Check whether `string` is a valid S3 URL using regexp"""
    if not isinstance(string, str):
        return False
    else:
        return REGEXP_S3_URL.match(string.strip())


@functools.lru_cache()
def parse_s3_url(url):
    """Parse S3 `url`, returning `endpoint` URL and `key`"""
    urlp = urlparse(url)
    scheme = urlp.scheme or "https"
    port = urlp.port or (80 if scheme == "http" else 443)
    s3_endpoint = f"{scheme}://{urlp.hostname}:{port}"
    s3_path = urlp.path
    return s3_endpoint, s3_path



"""DCOR feature handling"""
from functools import lru_cache
import numbers

import numpy as np

from ... import definitions as dfn


class DCORNonScalarFeature:
    """Helper class for accessing non-scalar features"""

    def __init__(self, feat, api, size):
        self.identifier = api.url + ":" + feat  # for caching ancillaries
        self.feat = feat
        self.api = api
        self._size = size

    def __iter__(self):
        for idx in range(len(self)):
            yield self[idx]

    def __getitem__(self, event):
        if not isinstance(event, numbers.Integral):
            # slicing!
            indices = np.arange(len(self))[event]
            trace0 = self._get_item(indices[0])
            # determine the correct shape from the first feature
            oshape = [len(indices)] + list(trace0.shape)
            output = np.zeros(oshape, dtype=trace0.dtype)
            # populate the output array
            for ii, evid in enumerate(indices):
                output[ii] = self._get_item(evid)
            return output
        else:
            return self._get_item(event)

    def __len__(self):
        return self._size

    @lru_cache(maxsize=100)
    def _get_item(self, event):
        data = self.api.get(query="feature", feat=self.feat, event=event)
        return np.asarray(data)


class DCORContourFeature(DCORNonScalarFeature):
    """Helper class for accessing contour data"""

    def __init__(self, feat, api, size):
        super(DCORContourFeature, self).__init__(feat, api, size)
        self.shape = (size, np.nan, 2)

    def __getitem__(self, event):
        if not isinstance(event, numbers.Integral):
            # We cannot use the original method, because contours
            # may have different sizes! So we return a list.
            indices = np.arange(len(self))[event]
            output = []
            # populate the output list
            for evid in indices:
                output.append(self._get_item(evid))
            return output
        else:
            return self._get_item(event)


class DCORImageFeature(DCORNonScalarFeature):
    """Helper class for accessing image data"""

    def __init__(self, feat, api, size):
        super(DCORImageFeature, self).__init__(feat, api, size)
        metadata = self.api.get(query="metadata")
        self.shape = (size,
                      metadata["imaging"]["roi size y"],
                      metadata["imaging"]["roi size x"])


class DCORTraceItem(DCORNonScalarFeature):
    """Helper class for accessing traces"""
    def __init__(self, feat, api, size, samples_per_event):
        super(DCORTraceItem, self).__init__(feat, api, size)
        self.shape = (size, samples_per_event)

    @lru_cache(maxsize=100)
    def _get_item(self, event):
        data = self.api.get(query="feature", feat="trace",
                            trace=self.feat, event=event)
        return np.asarray(data)


class DCORTraceFeature:
    """Helper class for accessing traces"""

    def __init__(self, api, size):
        self.identifier = api.url + ":traces"
        self.api = api
        self._size = size
        metadata = self.api.get(query="metadata")
        self._samples_per_event = metadata["fluorescence"]["samples per event"]
        self.traces = api.get(query="trace_list")
        self._trace_objects = {}

        self.shape = (len(self.traces),
                      size,
                      self._samples_per_event
                      )

    def __contains__(self, key):
        return key in self.traces

    def __getitem__(self, trace):
        if trace in self.traces:
            if trace not in self._trace_objects:
                self._trace_objects[trace] = DCORTraceItem(
                    trace, self.api, self._size, self._samples_per_event)
            return self._trace_objects[trace]
        else:
            raise KeyError(f"trace '{trace}' not found!")

    def __len__(self):
        return len(self.traces)

    def keys(self):
        return self.traces


class FeatureCache:
    """Download and cache (scalar only) features from DCOR"""

    def __init__(self, api, size):
        self.api = api
        self._features = self.api.get(query="feature_list")
        self._size = size
        self._scalar_cache = {}
        self._nonsc_features = {}

    def __contains__(self, key):
        return key in self._features

    def __getitem__(self, key):
        # user-level checking is done in core.py
        assert dfn.feature_exists(key)
        if key not in self._features:
            raise KeyError(f"Feature '{key}' not found!")

        if key in self._scalar_cache:
            return self._scalar_cache[key]
        elif dfn.scalar_feature_exists(key):
            # download the feature and cache it
            feat = np.asarray(self.api.get(query="feature", feat=key))
            self._scalar_cache[key] = feat
            return feat
        elif key == "contour":
            if key not in self._nonsc_features:
                self._nonsc_features[key] = DCORContourFeature(key, self.api,
                                                               self._size)
            return self._nonsc_features[key]
        elif key == "trace":
            if "trace" not in self._nonsc_features:
                self._nonsc_features["trace"] = DCORTraceFeature(self.api,
                                                                 self._size)
            return self._nonsc_features["trace"]
        elif key in ["image", "image_bg", "mask"]:
            self._nonsc_features[key] = DCORImageFeature(key, self.api,
                                                         self._size)
            return self._nonsc_features[key]
        else:
            raise ValueError(f"No handler in dclab for DCOR feature '{key}'!")

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def keys(self):
        return self._features



"""Test command-line interface dclab-condense"""
from dclab import cli, new_dataset

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


def test_check_suffix_disabled_condense():
    path_in_o = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_in = path_in_o.with_suffix("")
    path_in_o.rename(path_in)
    assert path_in.suffix == ""
    with pytest.raises(ValueError, match="Unsupported file type"):
        cli.condense(path_in=path_in,
                     path_out=path_in.with_name("condensed.rtdc"))
    # but this should work:
    cli.condense(path_in=path_in,
                 path_out=path_in.with_name("condensed2.rtdc"),
                 check_suffix=False)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    cli.condense(path_out=path_out, path_in=path_in)
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-condense" in dsj.logs
        assert len(dsj)
        assert len(dsj) == len(ds0)
        for feat in dsj.features:
            assert np.all(dsj[feat] == ds0[feat])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_no_ancillary_features():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    with h5py.File(path_in, "a") as h5:
        del h5["events/area_um"]

    cli.condense(path_out=path_out, path_in=path_in, ancillaries=False)
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "volume" in ds0.features
        assert "volume" not in ds0.features_innate
        assert "contour" not in dsj.features
        assert "image" not in dsj.features
        assert "volume" not in dsj.features, "expensive feature not computed"
        assert "area_um" not in ds0.features_innate, "sanity check"
        assert "area_um" in ds0.features
        assert "area_um" in dsj.features_innate, "cheap feature available"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_no_ancillary_features_control():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    with h5py.File(path_in, "a") as h5:
        del h5["events/area_um"]

    cli.condense(path_out=path_out, path_in=path_in)  # defaults to True
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "volume" in ds0.features
        assert "volume" not in ds0.features_innate
        assert "volume" in dsj.features
        assert "area_um" not in ds0.features_innate, "sanity check"
        assert "area_um" in ds0.features
        assert "area_um" in dsj.features_innate


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_condense_wo_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    with h5py.File(path_in, "a") as h5:
        del h5["logs"]
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("condensed.rtdc")

    cli.condense(path_out=path_out, path_in=path_in)
    with new_dataset(path_out) as ds:
        assert len(ds.logs) == 2
        assert "dclab-condense" in ds.logs
        assert "dclab-condense-warnings" in ds.logs



import functools

import numpy as np


class DCORTables:
    def __init__(self, api):
        self.api = api

    def __getitem__(self, key):
        return self._tables[key]

    def __len__(self):
        return len(self._tables)

    def keys(self):
        return self._tables.keys()

    @property
    @functools.lru_cache()
    def _tables(self):
        table_data = self.api.get(query="tables")
        # assemble the tables
        tables = {}
        for key in table_data:
            columns, data = table_data[key]
            ds_dt = np.dtype({'names': columns,
                              'formats': [np.float64] * len(columns)})
            tab_data = np.asarray(data)
            rec_arr = np.rec.array(tab_data, dtype=ds_dt)
            tables[key] = rec_arr

        return tables



"""Create .rtdc files with scalar-only features"""
import argparse
import pathlib
import warnings

import h5py
import hdf5plugin

from ..rtdc_dataset import new_dataset, rtdc_copy, RTDCWriter
from .. import util
from .._version import version

from . import common


def condense(path_out=None, path_in=None, ancillaries=True,
             check_suffix=True):
    """Create a new dataset with all (ancillary) scalar-only features"""
    cmp_kw = hdf5plugin.Zstd(clevel=5)
    if path_out is None or path_in is None:
        parser = condense_parser()
        args = parser.parse_args()
        path_in = args.input
        path_out = args.output
        ancillaries = not args.no_ancillaries

    allowed_input_suffixes = [".rtdc", ".tdms"]
    if not check_suffix:
        allowed_input_suffixes.append(pathlib.Path(path_in).suffix)

    path_in, path_out, path_temp = common.setup_task_paths(
        path_in, path_out, allowed_input_suffixes=allowed_input_suffixes)

    cmd_dict = {}

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        with new_dataset(path_in) as ds, h5py.File(path_temp, "w") as hc:
            # If we have an input HDF5 file, then we might readily copy most
            # of the features over using rtdc_copy. If we have a .tdms file,
            # then we have to go the long path.
            if ds.format == "hdf5":
                rtdc_copy(src_h5file=ds.h5file,
                          dst_h5file=hc,
                          features="scalar",
                          include_logs=True,
                          include_tables=True,
                          meta_prefix="")

            hc.require_group("logs")

            # scalar features
            feats_sc = ds.features_scalar
            # loaded (computationally cheap) scalar features
            feats_sc_in = [f for f in ds.features_loaded if f in feats_sc]
            # ancillary features
            feats_sc_anc = list(set(feats_sc) - set(feats_sc_in))

            cmd_dict["features_original_innate"] = ds.features_innate

            if ancillaries:
                features = feats_sc
                cmd_dict["features_computed"] = feats_sc_anc
                if feats_sc_anc:
                    print("Computing ancillary features:",
                          " ".join(feats_sc_anc))
                else:
                    print("No ancillary features to compute.")
            else:
                features = feats_sc_in

            # command log
            logs = {"dclab-condense": common.get_command_log(
                paths=[path_in], custom_dict=cmd_dict)}

            # rename old dclab-condense logs
            for lkey in ["dclab-condense", "dclab-condense-warnings"]:
                if lkey in hc["logs"]:
                    # This is cached, so no worry calling it multiple times.
                    md55m = util.hashfile(path_in, count=80)
                    # rename
                    hc["logs"][f"{lkey}_{md55m}"] = hc["logs"][lkey]
                    del hc["logs"][lkey]

            # Write all remaining scalar features to the file
            # (these are *all* features for .tdms data).
            with RTDCWriter(hc,
                            mode="append",
                            compression_kwargs=cmp_kw,
                            ) as hw:
                for feat in features:
                    if feat not in hc["events"]:
                        hw.store_feature(feat=feat, data=ds[feat])

        # warnings log
        if w:
            logs["dclab-condense-warnings"] = common.assemble_warnings(w)

    # Write log file
    with RTDCWriter(path_temp, compression_kwargs=cmp_kw) as hw:
        for name in logs:
            hw.store_log(name, logs[name])

    # Finally, rename temp to out
    path_temp.rename(path_out)


def condense_parser():
    descr = "Reduce an RT-DC measurement to its scalar-only features " \
            + "(i.e. without `contour`, `image`, `mask`, or `trace`). " \
            + "All available ancillary features are computed."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.tdms or .rtdc file)')
    parser.add_argument('output', metavar="OUTPUT", type=str,
                        help='Output path (.rtdc file)')
    parser.add_argument('--no-ancillary-features',
                        dest='no_ancillaries',
                        action='store_true',
                        help='Do not compute expensive ancillary features '
                             'such as volume'
                        )
    parser.add_argument('--version', action='version',
                        version=f'dclab-condense {version}')
    return parser
